@@ -1,20 +1,21 @@
 #!/usr/bin/env/python
 
 import json
+import math
 import pandas as pd
 import numpy as np
 import requests
 
 from audl.stats.endpoints._base import Endpoint
-from audl.stats.static import players
 from audl.stats.endpoints.playerprofile import PlayerProfile
 
 from audl.stats.library.game_event import GameEventSimple, GameEventLineup, GameEventReceiver
+from audl.stats.static.utils import get_quarter, get_throw_type, get_throwing_distance
 
 
-from audl.stats.library.parameters import quarters_clock_dict
-from audl.stats.library.parameters import HerokuPlay
-from audl.stats.library.parameters import team_roster_columns_name
+#  from audl.stats.library.parameters import quarters_clock_dict
+#  from audl.stats.library.parameters import team_roster_columns_name
+#  from audl.stats.library.parameters import HerokuPlay
 
 #  https://audl-stat-server.herokuapp.com/stats-pages/game/2022-06-11-TOR-MTL
 
@@ -186,45 +187,10 @@ class GameStats(Endpoint):
         
         # calculate quarter columns
         scores['quarter'] = scores['scoring_time'].apply(
-                lambda x: self._get_quarter(x))
+                lambda x: get_quarter(x))
 
         return scores
 
-
-    @staticmethod
-    def _get_quarter(scoring_time):
-        """ 
-        Function that returns quarter based on scoring time
-
-        Remark: Games are timed with four-quarters of 12 minutes each, 
-            including a 15-minute halftime. If the score is tied, a five 
-            minutes overtime period is played. If the score remains tied after
-            overtime, a second overtime is played in which the first team 
-            to score wins.
-
-        Parameters 
-        ----------
-        scoring_time: int
-            time when point was scored in second
-
-        Returns
-        -------
-        quarter: string 
-            ex: 'Q1', 'Q2', 'Q3', 'Q4', 'OT1', 'OT2'
-            
-        Examples
-        --------
-        >>> GameStats()._get_quarter(456)
-
-        """
-        quarter_end = { 'Q1': 720, 'Q2': 1440, 'Q3': 2160, 'Q4': 2880, 'OT1': 3180 }
-
-        for quarter, end_time in quarter_end.items():
-            if scoring_time < end_time:
-                return quarter
-
-        return 'OT2'
-
         
     def get_scores(self):
         """ 
@@ -534,7 +500,6 @@ class GameStats(Endpoint):
         >>> GameStats('2022-06-11-TB-ATL').get_lineup_by_points()
         """
         raise NotImplementedError("This function hasn't been implemented yet!")
-        pass
 
     def get_point_results(self, point_events):
         """ 
@@ -572,9 +537,8 @@ class GameStats(Endpoint):
 
         """
         raise NotImplementedError("This function hasn't been implemented yet!")
-        pass
 
-    def get_events(self):
+    def get_events_sequential(self):
         """ 
         Function that return the event of each points in sequential order
 
@@ -595,8 +559,326 @@ class GameStats(Endpoint):
 
         """
         raise NotImplementedError("This function hasn't been implemented yet!")
-        pass
 
+    def get_events(self):
+        """ 
+        Function that fetch all events for all points for each team
+        """
+
+        # retrieve events from json
+        home_events = json.loads(self.json['tsgHome']['events'])
+        away_events = json.loads(self.json['tsgAway']['events'])
+
+        # get events by points
+        home_points = self._get_team_events_by_points(home_events)
+        away_points = self._get_team_events_by_points(away_events)
+
+        events_dict = {'homeEvents': home_points, 'awayEvents': away_points}
+
+        return events_dict
+
+        
+    def _get_team_events_by_points(self, events):
+        """ 
+        Function that return a list of dict with all points events
+
+        Parameters
+        ----------
+        events: json dict
+            self.json['tsgHome']['events'] or self.json['tsgAway']['events']
+
+        Examples
+        --------
+        >>> 
+        >>> [{'point': 0, 'events': []},
+        >>>  {'point': 1,
+        >>>    'events': [{'t': 1, 'l': [9246, 9237, 9323, 9262, 9241, 9568, 9242]},
+        >>>     {'t': 20, 'r': 9262, 'x': 12.89, 'y': 18.93},
+        >>>     {'t': 20, 'r': 9246, 'x': 4, 'y': 33.13},
+        >>>     {'t': 20, 'r': 9323, 'x': 20.47, 'y': 47.18},
+        >>>     {'t': 20, 'r': 9262, 'x': 0.94, 'y': 43.83},
+        >>>     ...
+
+        """
+        point_played = 0
+        all_events, point_events = [], []
+        for i, event in enumerate(events):
+            event_type = int(event['t'])
+            if event_type not in [1,2]:
+                point_events.append(event)
+            else:
+                all_events.append({'point': point_played, 'events': point_events})
+                point_events = []
+                point_played += 1
+
+        return all_events
+        
+    def get_throw_selection(self):
+        """ 
+        Function that count throws types for all players
+
+        Returns
+        -------
+        df: pandas dataframe
+            players x types of throws (pass, dump, huck, swing, throwaway, dish)
+
+        Example
+        -------
+
+        """
+        # get events
+        home_events = json.loads(self.json['tsgHome']['events'])
+        away_events = json.loads(self.json['tsgAway']['events'])
+         
+
+        # get players id
+        players = self.get_players_metadata()
+        players = players[['id', 'player.first_name', 'player.last_name', 
+            'player.ext_player_id']]
+
+        # initialize df
+        type_of_throws = ['pass', 'huck', 'swing', 'dump', 'dish', 'throwaway', 'drop']
+        freq = [[0 for _ in range(len(type_of_throws))] for _ in range(len(players))]
+        df = pd.DataFrame(freq, columns=type_of_throws)
+        df['player'] = list(players['player.ext_player_id'])
+
+        # reorder columns
+        cols = df.columns.tolist()
+        cols = cols[-1:] + cols[:-1]
+        df = df[cols]
+
+        # count throw freq
+        self._count_throw_frequency(home_events, players, df)
+        self._count_throw_frequency(away_events, players, df)
+
+
+        # count total
+        df['total'] = df.sum(axis=1)
+
+        return df
+
+    def _count_throw_frequency(self, events, players, df):
+        """ 
+        Helper function for get_throw_selection() that count throw frequency
+        for a team 
+        """
+        # count throw frequency
+        x1, y1, player1 = None, None, None
+        for i, event in enumerate(events):
+            event_type = int(event['t'])
+            if event_type in [8, 19, 20, 22]: # event has x and y
+                x2, y2 = event['x'], event['y']
+                if event_type != 8:
+                    player2 = int(event['r'])
+                if x1 and y1 and player1:
+                    # get player
+                    player_id = players[players['id'] == player1]['player.ext_player_id'].tolist()[0]
+
+                    # get distance and throw selection
+                    #  dist = get_throwing_distance(x1, y1, x2, y2)
+                    throw = get_throw_type(x1, y1, x2, y2, event_type)
+                    #  print(player_id, throw, dist)
+
+                    # increment player throw selection
+                    df.loc[df['player'].isin([player_id]), throw] += 1
+                x1, y1 = x2, y2
+                if event_type != 8:
+                    player1 = player2
+            else: 
+                x1, y1, player1 = None, None, None
+
+
+    def get_thrower_receiver_count(self, is_home):
+        """ 
+        Function that returns the thrower-receiver count only if throw was 
+        successful. Useful to view players chemistry
+
+        Parameters
+        ----------
+        is_home: bool
+
+        Returns
+        -------
+        df: pandas dataframe
+            players x players: +1 if thrower attempted a pass to receiver
+
+        Example
+        -------
+        """
+        if is_home: 
+            events = json.loads(self.json['tsgHome']['events'])
+            players = pd.json_normalize(self.json['rostersHome'])
+        else: 
+            events = json.loads(self.json['tsgAway']['events'])
+            players = pd.json_normalize(self.json['rostersAway'])
+
+        # initialize df
+        list_players = list(players['player.ext_player_id'])
+        freq = [[0 for _ in range(len(list_players))] for _ in range(len(list_players))]
+        df = pd.DataFrame(freq, columns=list_players)
+        df['player'] = list_players
+
+        # reorder columns
+        cols = df.columns.tolist()
+        cols = cols[-1:] + cols[:-1]
+        df = df[cols]
+
+        # count thrower-receiver freq 
+        self._count_thrower_receiver(events, players, df)
+
+        # count total throws
+        df['total'] = df.sum(axis=1)
+
+        return df
+
+    def _count_thrower_receiver(self, events, players, df):
+        """ 
+        Helper function for get_thrower_receiver_count() to count thrower and
+        receiver for team event
+        """
+        player1 = None
+        for i, event in enumerate(events):
+            event_type = int(event['t'])
+            if event_type in [20, 22]: # removed 19 because dropped
+                player2 = int(event['r'])
+                if player1:
+                    # get player
+                    player1_id = players[players['id'] == player1]['player.ext_player_id'].tolist()[0]
+                    player2_id = players[players['id'] == player2]['player.ext_player_id'].tolist()[0]
+
+                    # increment thrower-receiver 
+                    df.loc[df['player'].isin([player1_id]), player2_id] += 1
+                player1 = player2
+            else: 
+                player1 = None
+
+    def get_lineup_frequency(self, is_home): 
+        """ 
+        Function that calculates the number of time a player is on the same 
+        line as teamate. Only counts starting lineup (ie starting on offense 
+        or defense, not after timeout)
+
+        Parameters
+        ----------
+        is_home: bool
+
+        Returns
+        -------
+        df: pandas dataframe
+            players x players: +1 if player i and j are on the same line
+        """
+        if is_home: 
+            events = json.loads(self.json['tsgHome']['events'])
+            players = pd.json_normalize(self.json['rostersHome'])
+        else: 
+            events = json.loads(self.json['tsgAway']['events'])
+            players = pd.json_normalize(self.json['rostersAway'])
+
+        # initialize df
+        list_players = list(players['player.ext_player_id'])
+        freq = [[0 for _ in range(len(list_players))] for _ in range(len(list_players))]
+        df = pd.DataFrame(freq, columns=list_players)
+        df['player'] = list_players
+
+        # reorder columns
+        cols = df.columns.tolist()
+        cols = cols[-1:] + cols[:-1]
+        df = df[cols]
+
+        # count lineup freq
+        self._count_lineup_frequency(events, players, df)
+
+        return df
+
+    def _count_lineup_frequency(self, events, players, df): 
+        """ 
+        Helper function for get_lineup_frequency()
+        """
+        for i, event in enumerate(events):
+            event_type = int(event['t']) 
+            if event_type in [1,2]:
+                l = event['l']
+                lineup = [players[players['id'] == int(player_id)]['player.ext_player_id'].tolist()[0] for player_id in l]
+                #  print(lineup)
+                for i in range(len(lineup)):
+                    for j in range(len(lineup)):
+                        if i != j:
+                            player1, player2 = lineup[i], lineup[j]
+                            df.loc[df['player'].isin([player1]), player2] += 1
+        
+
+    def get_teamates_selection(self, player_id, is_home):
+        """ 
+        For a given thrower, return throw selection for each teamates
+
+        Parameters
+        ----------
+        is_home: bool
+            True if player is in home team, else false
+        player_id: str
+            player external id
+
+        Returns
+        -------
+        df: pandas dataframe
+            teamates x throwing_type
+
+        Example
+        -------
+        >>> game.get_thrower_receiver_selection('cbrock', True)
+        """
+        if is_home: 
+            events = json.loads(self.json['tsgHome']['events'])
+            players = pd.json_normalize(self.json['rostersHome'])
+        else: 
+            events = json.loads(self.json['tsgAway']['events'])
+            players = pd.json_normalize(self.json['rostersAway'])
+
+        # check if player_id in players list
+        list_players = list(players['player.ext_player_id'])
+        if player_id not in list_players:
+            raise ValueError("player_id doesn't exist. Please check!")
+
+        # initialize dataframe
+        type_of_throws = ['pass', 'huck', 'swing', 'dump', 'dish', 'throwaway', 'drop']
+        selection = [[0 for _ in range(len(type_of_throws))] for _ in range(len(list_players))]
+        df = pd.DataFrame(selection, columns=type_of_throws)
+        df['player'] = list_players
+
+        # reorder columns
+        cols = df.columns.tolist()
+        cols = cols[-1:] + cols[:-1]
+        df = df[cols]
+
+        # count teamates selection
+        self._count_teamate_selection(events, players, player_id, df)
+
+        # total throws by teamates
+        df['total'] = df.sum(axis=1)
+
+        return df
+        
+    def _count_teamate_selection(self, events, players, thrower, df): 
+        """ 
+        Helper Function for get_teamates_selection()
+        """
+        x1, y1, player1 = None, None, None
+        for i, event in enumerate(events):
+            event_type = int(event['t'])
+            if event_type in [19, 20, 22]: # event has x and y
+                x2, y2 = event['x'], event['y']
+                player2 = int(event['r'])
+                player_id = players[players['id'] == player2]['player.ext_player_id'].tolist()[0]
+                if x1 and y1 and player1 == thrower:
+                    # get distance and throw selection
+                    throw = get_throw_type(x1, y1, x2, y2, event_type)
+
+                    # increment player throw selection
+                    df.loc[df['player'].isin([player_id]), throw] += 1
+                x1, y1 = x2, y2
+                player1 = player_id
+            else: 
+                x1, y1, player1 = None, None, None
 
         
     def print_team_events(self, is_home): 
@@ -642,7 +924,10 @@ class GameStats(Endpoint):
                     receiver = players[players['id'] == int(row['r'])]['player.ext_player_id'].tolist()[0]
                 except: 
                     receiver = 'NaN'
-                print(f"t: {t}; r: {receiver}")
+                if t in [3,19,20,22]:
+                    print(f"t: {t}; r: {receiver}; x: {row['x']}; y: {row['y']}")
+                else: 
+                    print(f"t: {t}; r: {receiver}")
             elif t in [14, 15, 42, 43]:
                 # print s
                 print(f"t: {t}; s: {row['s']}")
